/* napster.c - all napster-related protocol stuff */

#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include "napster.h"
#include "gnapster.h"
#include "network.h"
#include "ui.h"
#include "upload.h"
#include "callbacks.h"
#include "queue.h"
#include "regexp.h"
#include "chan.h"
#include "text.h"
#include "xtext.h"
#include "utils.h"
#include "connect.h"

#include "commands.h"

#include "debug.h"

extern ConnInfo conn_info;
extern MOTDTab motd_tab;
extern SearchTab search_tab;
extern BrowseTab browse_tab;
extern ConsoleTab console_tab;
extern UserInfo user_info;
extern ListChanDlg list_chan_dlg;

extern DownloadTab download_tab;
extern UploadTab upload_tab;

extern GList *ignore_list;
extern GList *signore_list;
extern GdkPixmap *speedgreen, *speedyellow, *speedred, *speedgray;
extern GdkBitmap *speedgreenb, *speedyellowb, *speedredb, *speedgrayb;

extern int nick_curr, nick_end;
extern GList *nick_history;

extern int na_err;

int last_dns = 0, browse_results = 0;
int flag = 0;

char *conn_table[] = {
   "Unknown", "14.4", "28.8", "33.6", "56k", "64k ISDN", "128k ISDN",
   "Cable", "DSL", "T1", "T3+", NULL
};


int napster_send(int sock, int cmd, char *fmt, ...) {
   Packet *p;
   va_list args;
   char *data = NULL;
   int x;
   
   if (cmd < 0 || sock < 0) 
     return -1;
   
   if (fmt) {
      va_start(args, fmt);
      data = d_strdup_vprintf(fmt, args);
      va_end(args);
   }
   
   p = d_calloc(sizeof(Packet));

   p->len = data ? strlen(data) : 0;
   p->cmd = cmd;
   
   /* if our host order is BIG_ENDIAN, we need to reverse the
    * bytes */
   p->len = BSWAP16(p->len);
   p->cmd = BSWAP16(p->cmd);
      
   x = send(sock, p, 4, 0);
   if (data && x > 0)
     x += n_send(sock, "%s", data);
   
   d_free(data);
   d_free(p);
   
   return x;
}

void napster_send_get(int sock, char *user, char *file, unsigned long int size) {
   n_send(sock, "GET");
   n_send(sock, "%s \"%s\" %lu", user, file, size);
}

void napster_send_send(int sock, char *user, char *file, unsigned long int size) {
   n_send(sock, "SEND");
   n_send(sock, "%s \"%s\" %lu", user, file, size);
}

int opennap_version(int major, int minor) {
   int ma, mi;
   unsigned short on_version;
   
   on_version = ((conn_info.state & OPENNAP_MASK) >> 16);

   if (!on_version)
     return 0;
   
   ma = on_version >> 8;
   mi = on_version & 0xff;
   
   if (ma > major || (ma == major && mi >= minor))
     return 1;
   
   return 0;
}

int is_opennap_server(char *data) {
   char *ver, *server, *server_ver;
   int ver_minor, ver_major;
   
   ver = next_arg(data, &data);
   server = next_arg(data, &data);
   server_ver = next_arg(data, &data);
   
   NA_ERR_HANDLE(0);

   if (strcmp(ver, "VERSION")) 
     return 0;
   
   if (strcmp(server, "opennap")) 
     return 0;
   
   ver_major = my_atoi(next_arg_full(server_ver, &server_ver, '.'));
   ver_minor = my_atoi(next_arg_full(server_ver, &server_ver, '.'));
   
   NA_RESET();
   
   conn_info.state |= (((ver_major & 0xff) << 8 | (ver_minor)) << 16);
     
   if (ver_minor < 0) 
     return 0;
   
   if (ver_major > 0 || (ver_major == 0 && ver_minor >= 12))
     return 1;
   
   return 0;
}

int user_ignored(char *user) {
   GList *ptr;
   char *nick;
   
   if (!user) 
     return 0;
   
   for(ptr=ignore_list; ptr; ptr=ptr->next) {
      nick = ptr->data;
      if (!nick)
	continue;
      
      if (!strcmp(nick, user))
	return 1;
   }
   
   return 0;
}

void napster_show_links(GtkCTreeNode *node, char *print_str) {
   char *buf;
   char print_str_local[80];
   
   if (print_str) 
     strcpy(print_str_local, print_str);
   else
     print_str_local[0] = '\0';
   
   if (!node)
     node = GTK_CTREE_NODE(GTK_CLIST(console_tab.links_map)->row_list);
   while(node) {
      buf = GTK_CELL_PIXTEXT(GTK_CTREE_ROW(node)->row.cell[0])->text;
      gnapster_text_insert(CURR, SYSTEM,
			   "%s", print_str_local);
      if (GTK_CTREE_ROW(node)->parent)
	gnapster_text_insert(CURR, MESSAGE, 
			     "%s", (GTK_CTREE_ROW(node)->sibling) ? "|-- " : "`-- ");
      gnapster_text_insert(CURR, MESSAGE, "%s\n", buf);
      if (GTK_CTREE_ROW(node)->children) {
	 if (GTK_CTREE_ROW(node)->parent)
	   strcat(print_str_local,
		  GTK_CTREE_ROW(node)->sibling ? "|   " :
		                                 "    ");
	 else
	   strcat(print_str_local, "    ");
	 
	 napster_show_links(GTK_CTREE_ROW(node)->children, print_str_local);
	 print_str_local[strlen(print_str_local) - strlen("    ")] = 0;
      }
      node = GTK_CTREE_ROW(node)->sibling;
   }
}

GtkCTreeNode *napster_find_link(GtkCTreeNode *node, char *s) {
   GtkCTreeNode *work;
   char *buf;
   
   if (!s)
     return NULL;
   
   if (!node)
     node = GTK_CTREE_NODE(GTK_CLIST(console_tab.links_map)->row_list);
   while(node) {
      buf = GTK_CELL_PIXTEXT(GTK_CTREE_ROW(node)->row.cell[0])->text;
      if (!strcmp(buf, s))
	return node;
      if (GTK_CTREE_ROW(node)->children &&
	  (work = napster_find_link(GTK_CTREE_ROW(node)->children, s)))
	return work;
      node = GTK_CTREE_ROW(node)->sibling;
   }
   
   return NULL;
}

void napster_handle_data(int c, char *data) {
   char temp_str[1024];
   
//   printf("%i\t%s\n", c, data);

   switch(c) {
    case NAPSTER_MESSAGE:
      g_snprintf(temp_str, sizeof(temp_str), "%s\n", data);
      if (!GTK_XTEXT(motd_tab.text)->text_first) {
	 if (is_opennap_server(data)) {
	    gtk_widget_show(search_tab.not_label);
	    gtk_widget_show(search_tab.not_entry);
	    break;
	 }
      }
      gnapster_text_insert(MOTD, MESSAGE, temp_str);
      break;
    case NAPSTER_ADMIN_MESSAGE:
	{
	   char *user, *message;
	   int c;
	   
	   user = next_arg(data, &data);
	   message = last_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();

	   if (!strcmp(user, user_info.user))
	     break;
	   
	   if ((c = chan_lookup("#admins")) == -1)
	     c = CONSOLE;
	   
	   hook_text_insert(c, MESSAGE, "wallop", "%s\4%s",
			    user, message);
	}
      break;
    case NAPSTER_GLOBAL_MESSAGE:
	{
	   char *user, *message;
	   
	   user = next_arg(data, &data);
	   message = last_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   hook_text_insert(CONSOLE, MESSAGE, "global", "%s\4%s",
			    user, message);
	}
      break;
    case NAPSTER_LOGIN_ACK:
      g_snprintf(temp_str, sizeof(temp_str), _("Connected (%s:%i): "),
		 conn_info.ip, conn_info.port);
      
      if (conn_info.status)
	d_free(conn_info.status);
      
      conn_info.status = d_strdup(temp_str);
      
      if (!searching() && !browsing())
	gnapster_update_status(conn_info.status, 0);
      
      if (!shared_list_handle(user_info.upload_dir, SUBMIT_LIST))
	autojoin_handle();
      
      break;
    case NAPSTER_STATUS_MESSAGE:
	{
	   int songs, users, gigs;
      
	   users = my_atoi(next_arg(data, &data));
	   songs = my_atoi(next_arg(data, &data));
	   gigs = my_atoi(next_arg(data, &data));
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   g_snprintf(temp_str, sizeof(temp_str), 
		      _("Connected (%s:%i): %i songs, %i users, %i %s"),
		      conn_info.ip, conn_info.port, songs, users, gigs,
		      (gigs == 1) ? "gig" : "gigs");
	   
	   if (!searching() && !browsing())
	     gnapster_update_status(temp_str, 0);
	   
	   d_free(conn_info.status);
	   conn_info.status = d_strdup(temp_str);
	}
      break;
    case NAPSTER_USER_MESSAGE:
	{
	   char *user, *message;
	   int index;
	   
	   user = next_arg(data, &data);
	   message = last_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   if (user_ignored(user)) 
	     break;
	   
/*	   if (!strncmp(message, "\1SEND ", 6)) {
	      Transfer *dcc;
	      char *ip_str;
	      unsigned long int ip = 0;
	      
              dcc = d_new(N_DOWNLOAD);
	      
	      next_arg(message, &message);
	      
	      dcc->user = d_strdup(next_arg(message, &message));
	      ip_str = next_arg(message, &message);
	      dcc->port = my_atoi(next_arg(message, &message));
	      dcc->file = d_strdup(next_arg(message, &message));
	      dcc->connection = 0;
	      dcc->id = d_strdup(next_arg(message, &message));
	      dcc->kbps = my_atoi(next_arg(message, &message));
	      dcc->status = 1;
	      
	      if (dcc->kbps < 0) {
		 transfer_free(dcc);
		 break;
	      }
	      
	      sscanf(ip_str, "%lu", &ip);
	      dcc->ip = BSWAP32(ip);
	      dcc->trunc_file = d_strdup(dcc->file);
	      
	      network_add_download(dcc);
	      napster_connect_user(dcc);
	      
	      break;
	   }*/
	   
	   if (user_info.conf[AUTOQUERY])
	     if (user_lookup(user) == -1)
	       do_query(NULL, user);

	   index = user_lookup(user);
	   if (index > 1)
	     hook_text_insert(index, MESSAGE, "public_msg", "%s\4%s",
			      user, message);
	   else
	     hook_text_insert(CURR, MESSAGE, "privmsg", "%s\4%s",
			      user, message);
	   
	   nick_history = add_nick_history(nick_history, user);
	}
      break;
    case NAPSTER_SERVER_STATS:
	{
	   unsigned long clients, files, t, uptime, users, numusers, memory;
	   unsigned short servers, gigs, channels;
	   
	   convert(next_arg(data, &data), "%lu", &clients);
	   convert(next_arg(data, &data), "%hu", &servers);
	   convert(next_arg(data, &data), "%lu", &users);
	   convert(next_arg(data, &data), "%lu", &files);
	   convert(next_arg(data, &data), "%hu", &gigs);
	   convert(next_arg(data, &data), "%hu", &channels);
	   convert(next_arg(data, &data), "%lu", &t);
	   convert(next_arg(data, &data), "%lu", &uptime);
	   convert(next_arg(data, &data), "%lu", &memory);
	   convert(next_arg(data, &data), "%lu", &numusers);
	   
	   NA_ERR_HANDLE_BREAK();

	   gnapster_text_insert(CURR, SYSTEM, "clients = %lu, servers = %hu, users = %lu, files = %lu, gigs = %hu, channels = %hu, time = %lu, uptime = %lu, memory = %lu, numusers = %hu\n",
				clients, servers, users, files, gigs,
				channels, t, uptime, memory, numusers);
	}
      break;
    case NAPSTER_BAN_LIST_ENTRY:
    case NAPSTER_CBAN_LIST_ENTRY:
	{
	   char *ip, *nick, *reason;
	   time_t ban_time;
	   
	   if (!flag)
	     gnapster_text_insert(CURR, SYSTEM, "%-20s %-15s %s\n",
				  "Ban", "Who", "Reason");
	   
	   flag = 1;
	   
	   ip = next_arg(data, &data);
	   nick = next_arg(data, &data);
	   reason = next_arg(data, &data);
	   convert(next_arg(data, &data), "%lu", &ban_time);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   gnapster_text_insert(CURR, SYSTEM, "%-20s %-15s %s\n",
				ip, nick, reason);
	}
      break;
    case NAPSTER_BAN_LIST:
      if (!flag)
	hook_text_insert(CURR, SYSTEM, "general_message", "%s",
			 "no global bans found");
      
      flag = 0;
      break;
/*    case NAPSTER_CBAN_LIST_ENTRY:
	{
	   char *ban, *who, *reason;
	   time_t ban_time;
	   
	   if (!flag)
	     gnapster_text_Insert(CURR, SYSTEM, "\2%-20s %-15s %s\n",
				  "Ban", "Who", "Reason");
	   flag = 1;
	   
	   ban = next_arg(data, &data);
	   who = next_arg(data, &data);
	   reason = next_arg(data, &data);
	   convert(next_arg(data, &data, "%lu", &ban_time));
	   if (ban_time < 0)
	     break;
	   
	   gnapster_text_insert(CURR, SYSTEM, "%-20s %-15s %s\n",
				ban, who, reason);
	}
      break;*/
    case NAPSTER_CBAN_LIST:
      if (!flag)
	hook_text_insert(CURR, SYSTEM, "general_message", "%s",
			 "no channel bans found");
      
      flag = 0;
      break;
    case NAPSTER_CHAN_JOIN:
	{
	   GtkCList *clist;
	   
	   append_console_tab(data, 0);
	   clist = chan_lookup_clist(data);
	   if (!clist) break;
	   
	   gtk_notebook_set_page(GTK_NOTEBOOK(console_tab.notebook),
				 chan_lookup(data) - 1);
	   gtk_clist_freeze(clist);
	}
      break;
    case NAPSTER_PART_CHAN:
	{
	   char *chan;
	   
	   chan = next_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   remove_console_tab(chan, 0);
	}
      break;
    case NAPSTER_CHAN_LIST_ENTRY:
	{
	   char *chan, *users, *topic;
	   Channel *nchan;
	   
	   chan = next_arg(data, &data);
	   users = next_arg(data, &data);
	   topic = last_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   if (!dialog_open(list_chan_dlg.window))
	     break;
	   
	   nchan = d_new(N_CHAN_INFO);

	   nchan->chan = d_strdup(chan);
	   nchan->users = my_atoi(users);
	   nchan->topic = d_strdup(topic);
	   
	   gnapster_clist_append(list_chan_dlg.clist, NULL, nchan,
				 nchan->chan, users, nchan->topic, NULL);
	   
	   gtk_clist_sort(GTK_CLIST(list_chan_dlg.clist));
	}
      break;
    case NAPSTER_CHAN_LIST:
      if (!dialog_open(list_chan_dlg.window))
	break;
      
      gtk_clist_thaw(GTK_CLIST(list_chan_dlg.clist));
      gtk_label_set_text(GTK_LABEL(list_chan_dlg.label), "");
      break;
    case NAPSTER_CHAN_JOINUSER:
    case NAPSTER_CHAN_PARTUSER:
	{
	   char *chan, *user, *sharing;
	   int link_type;
	   GtkCList *clist;
	   
	   chan = next_arg(data, &data);
	   user = next_arg(data, &data);
	   sharing = next_arg(data, &data);
	   link_type = my_atoi(next_arg(data, &data));
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   if (!user_info.conf[HIDE_JOINPART])
	     hook_text_insert(chan_lookup(chan),
			      (c == NAPSTER_CHAN_JOINUSER) ? SYSTEM_P : SYSTEM_N,
			      (c == NAPSTER_CHAN_JOINUSER) ? "channel_join" : "channel_part",
			      "%s\4%s\4%s\4%s", user, chan,
			      napster_conn_as_str(link_type), sharing);
	   
	   clist = chan_lookup_clist(chan);
	   if (!clist) break;
	   
	   if (c == NAPSTER_CHAN_JOINUSER) {
	      User *nusr;
	      
	      nusr = d_new(N_USER);

	      nusr->user = d_strdup(user);
	      nusr->shares = my_atoi(sharing);
	      nusr->conn = link_type;
	      
	      gnapster_clist_append(GTK_WIDGET(clist), NULL, nusr, nusr->user, 
				    sharing, napster_conn_as_str(link_type), 
				    NULL);
	      gtk_clist_sort(clist);
	   } else if (c == NAPSTER_CHAN_PARTUSER) {
	      int i, x = g_list_length(clist->row_list);
	      char *buf;
	      for(i=0; i<x; i++) {
		 gtk_clist_get_text(clist, i, 0, &buf);
		 if (!strcmp(buf, user)) {
		    gtk_clist_remove(clist, i);
		    gtk_clist_sort(clist);
		    break;
		 }
	      }
	   }
	}
      break;
    case NAPSTER_CHAN_USERLIST:
	{
	   char *chan, *user, *share;
	   int conn;
	   User *nusr;
	   GtkCList *clist;
	   
	   chan = next_arg(data, &data);
	   user = next_arg(data, &data);
	   share = next_arg(data, &data);
	   conn = my_atoi(next_arg(data, &data));
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   clist = chan_lookup_clist(chan);
	   
	   if (!clist) 
	     break;

	   nusr = d_new(N_USER);

	   nusr->user = d_strdup(user);
	   nusr->shares = my_atoi(share);
	   nusr->conn = conn;

	   gnapster_clist_append(GTK_WIDGET(clist), NULL, nusr, nusr->user, 
				    share, napster_conn_as_str(conn),
				    NULL);
	}
      break;
    case NAPSTER_CHAN_USERLIST_END:
	{
	   char *chan;
	   GtkCList *clist;
	   
	   chan = next_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   clist = chan_lookup_clist(chan);
	   if (!clist) 
	     break;
	   
	   gtk_clist_sort(clist);
	   gtk_clist_thaw(clist);
	}
      break;
    case NAPSTER_CHAN_TOPIC:
	{
	   char *chan, *topic;

	   chan = next_arg(data, &data);
	   topic = last_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   hook_text_insert(chan_lookup(chan), SYSTEM, "channel_topic",
			    "%s\4%s", chan, topic);
	}
      break;
    case NAPSTER_CHAN_MESSAGE_IN:
	{
	   char *chan, *user, *message;
	   
	   chan = next_arg(data, &data);
	   user = next_arg(data, &data);
	   message = last_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   if (user_ignored(user)) 
	     break;
	   
	   if (strcmp(user, user_info.user)) {
	      char *al1, *al2;

	      al1 = d_strdup(all_lower(message));
	      al2 = d_strdup(all_lower(user_info.user));
	      if (!al1 || !al2)
		break;
	      	      
	      if (strstr(al1, al2))
		hook_text_insert(chan_lookup(chan), MESSAGE, "public_msg_highlight",
				 "%s\4%s", user, message);
	      else
		hook_text_insert(chan_lookup(chan), MESSAGE, "public_msg",
				 "%s\4%s", user, message);
	      
	      d_free(al1);
	      d_free(al2);
	   }
	}
      break;
    case NAPSTER_LINKS:
	{
	   char *host1, *host2, *text[2], *hops;
	   int port1, port2;
	   GtkCTreeNode *parent;
	   
	   if (!data || !data[0]) {
	      if (console_tab.links_map) {
		 napster_show_links(NULL, NULL);
		 gtk_widget_destroy(console_tab.links_map);
	      } else
		hook_text_insert(CURR, SYSTEM, "general_message", "%s",
				 "no links found");
	      console_tab.links_map = NULL;
	      break;
	   }
	   
	   host1 = next_arg(data, &data);
	   port1 = my_atoi(next_arg(data, &data));
	   host2 = next_arg(data, &data);
	   port2 = my_atoi(next_arg(data, &data));
	   hops = next_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();

	   if (!console_tab.links_map)
	     console_tab.links_map = gtk_ctree_new(2, 0);
	   
	   parent = napster_find_link(NULL, host1);
	   if (!parent) {
	      text[0] = host1;
	      text[1] = "-1";
	      text[2] = NULL;
	      gtk_ctree_insert_node(GTK_CTREE(console_tab.links_map),
				    NULL, NULL, text, 0, NULL, NULL, NULL,
				    NULL, 0, 0);
	      
	      /* Look again now */
	      parent = napster_find_link(NULL, host1);
	   }
	   
	   text[0] = host2;
	   text[1] = hops;
	   text[2] = NULL;
	   
	   gtk_ctree_insert_node(GTK_CTREE(console_tab.links_map),
				 parent, NULL, text, 0, NULL, NULL, NULL, 
				 NULL, 0, 0);
	}
      break;
    case NAPSTER_WHOIS_REPLY:
	{
	   char *nick, *account, *channels, *status, *client, *host, *email;
	   int online_time, sharing, downloads, uploads, connection;
	   int t_uploads, t_downloads, l_port, d_port;
	   
	   nick = next_arg(data, &data);
	   account = next_arg(data, &data);
	   online_time = my_atoi(next_arg(data, &data));
	   channels = next_arg(data, &data);
	   status = next_arg(data, &data);
	   sharing = my_atoi(next_arg(data, &data));
	   downloads = my_atoi(next_arg(data, &data));
	   uploads = my_atoi(next_arg(data, &data));
	   connection = my_atoi(next_arg(data, &data));
	   client = next_arg(data, &data);
	   t_uploads = my_atoi(next_arg(data, &data));
	   t_downloads = my_atoi(next_arg(data, &data));
	   host = next_arg(data, &data);
	   l_port = my_atoi(next_arg(data, &data));
	   d_port = my_atoi(next_arg(data, &data));
	   email = next_arg(data, &data);
	   
	   NA_RESET();
	   
	   if (channels && *channels == ' ')
	     channels++;
	   
	   if (client && !email)
	     hook_text_insert(CURR, MESSAGE, "whois", "%s\4%s\4%i m %i s\4%s\4%s\4%i\4%i\4%i\4%s\4%s",
			      nick, account, online_time / 60, 
			      online_time % 60, channels, status, sharing, 
			      downloads, uploads,
			      napster_conn_as_str(connection), client);
	   else if (client && email)
	     hook_text_insert(CURR, MESSAGE, "whois_admin", "%s\4%s\4%i m %i s\4%s\4%s\4%i\4%i\4%i\4%s\4%s\4%i\4%i\4%s\4%i\4%i\4%s",
			      nick, account, online_time / 60, 
			      online_time % 60, channels, status, sharing, 
			      downloads, uploads,
			      napster_conn_as_str(connection), client,
			      t_uploads, t_downloads, host, l_port, d_port,
			      email);
	}
      break;
    case NAPSTER_WHOIS_OFFLINE:
	{
	   char *user, *account;

	   user = next_arg(data, &data);
	   account = next_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();

	   hook_text_insert(CURR, SYSTEM, "whois_failure", "%s",
			    user);
	}
      break;
    case NAPSTER_UPLOAD_REQUEST:
	{
	   char *user, *file, *trunc_file;
	   
	   user = next_arg(data, &data);
	   file = next_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();

	   trunc_file = strrchr(file, '\\');
	   if (!trunc_file) 
	     break;
	   
	   trunc_file++;

	   if (user_ignored(user))
	     break;

	   if (!upload_accepted(user, file, trunc_file))
	     break;
	   
	   napster_send(conn_info.sock, NAPSTER_ACCEPT_UPLOAD_REQUEST,
			"%s \"%s\"", user, file);
	}
      break;
    case NAPSTER_SERVER_MESSAGE:
	{
	   GList *ptr;
	   char *str;
	   int match = 0, c;
	   
	   for(ptr=signore_list; ptr; ptr=ptr->next) {
	      str = ptr->data;
	      if (str) {
		 if (wild_match(str, data))
		   match = 1;
	      }
	      if (match)
		break;
	   }
	   
	   if (!strncmp(data, "duplicate file", strlen("duplicate file")))
	     match = 1;
	   
	   c = wild_match("You * killed *", data) ? CONSOLE : CURR;
	   
	   if (!match)
	     hook_text_insert(c, SYSTEM, "server_msg", "%s", data);
	   
	   match = 0;
	   if (wild_match("User % is not * online*", data))
	     match = 1;
	   
	   if (match)
	     end_search(NULL);
	}
      break;
    case NAPSTER_RESUME_RESULT:
	{
	   Search *resume;
	   
	   resume = gnapster_parse_resume(data);
//	   gnapster_fill_resume_path(resume);
	   
	   gnapster_clist_append(search_tab.clist, search_tab.vbox,
				 resume, resume->trunc_file, 
				 napster_conn_as_str(resume->connection),
				 resume->user, "", "", NULL);
	}
      break;
    case NAPSTER_RESUME_END:
      end_search("No resume entries found");

      break;
    case NAPSTER_SEARCH_RESULT:
	{
	   Search *search;
	   char temp_str2[10], temp_str3[5];
	   int seconds, row;
	   GdkPixmap *pixmap = NULL;
	   GdkBitmap *bitmap = NULL;
	   Transfer *t;
	   
	   if (cancels())
	     break;
	   
	   search = gnapster_parse_search(data);
	   if (!search) 
	     break;
	   
	   seconds = (search->kbps) ?
	     (search->size / ((search->kbps * 1000) / 8)) : 0;
	   
	   g_snprintf(temp_str, sizeof(temp_str), "%.02fM", ((float)search->size / 1024) / 1024);
	   g_snprintf(temp_str2, sizeof(temp_str2), "%i:%02i",
		      seconds / 60, seconds % 60);
	   g_snprintf(temp_str3, sizeof(temp_str3), "%i", search->kbps);
	   row = gnapster_clist_append(search_tab.clist, search_tab.vbox,
				       search, search->trunc_file,
				       napster_conn_as_str(search->connection),
				       search->user, temp_str, temp_str2, temp_str3, NULL);
	   
	   if (search->connection > 10 || search->connection < 1)
	     pixmap = speedgray, bitmap = speedgrayb;
	   else if (search->connection <= 4)
	     pixmap = speedred, bitmap = speedredb;
	   else if (search->connection <= 6)
	     pixmap = speedyellow, bitmap = speedyellowb;
	   else if (search->connection <= 10)
	     pixmap = speedgreen, bitmap = speedgreenb;
	   else 
	     pixmap = speedgray, bitmap = speedgrayb;
	   
	   gtk_clist_set_pixtext(GTK_CLIST(search_tab.clist), row, 0,
				 search->trunc_file, 3, pixmap, bitmap);
	   
	   /* find out if this result is capable of being resumed... */
	   t = d_new(TRANSFER);
	   
	   t->size = search->size;
	   t->id = d_strdup(search->id);
	   
	   check_resume(t);
	   
	   if (t->resume_path)
	     set_green_style(search_tab.clist, row, 3);
	   
	   transfer_free(t);
	}
      break;
    case NAPSTER_END_SEARCH:
      if (!cancels())
	end_search("Nothing found for '%s'");
      else
	cancels_inc(-1);
      
      break;
    case NAPSTER_OPENNAP_BROWSE_RESULT:
	{
	   Search *browse;
	   char *buf, *user, *path, id[41];

	   user = next_arg(data, &data);
	   path = next_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   for(;;) {
	      browse = gnapster_parse_browse(data, &data, 1);
	      if (!browse) 
		break;
	      
	      memset(&id, '0', sizeof(id));
	      id[40] = 0;
	      
	      if (!strcmp(browse->id, "0")) {
		 d_free(browse->id);
		 browse->id = d_strdup(id);
	      }
	      
	      buf = d_strdup_printf(DEBUG "%s\\%s", path,
				    browse->trunc_file);
	      d_free(browse->file);
	      browse->file = buf;
	      
	      browse->user = d_strdup(user);
	      
	      browse_results++;
	      
	      browse_tab_add(browse);
	   }
	}
      break;
    case NAPSTER_OPENNAP_BROWSE:
	{
	   char *results;
	   
	   gnapster_update_status(conn_info.status, 0);
	   
	   gtk_entry_set_text(GTK_ENTRY(browse_tab.entry), "");
	   gtk_entry_set_editable(GTK_ENTRY(browse_tab.entry), 1);
	   
	   gtk_ctree_sort_recursive(GTK_CTREE(browse_tab.ctree), NULL);
	   gtk_clist_thaw(GTK_CLIST(browse_tab.ctree));
	   
	   if (g_list_length(GTK_CLIST(browse_tab.ctree)->row_list) == 0)
	     gnapster_clist_append(browse_tab.ctree, browse_tab.vbox, NULL,
				   "No mp3s found.", "", "", "", "", NULL);
	   
	   results = d_strdup_printf(DEBUG "%i files", browse_results);
	   gtk_clist_set_text(GTK_CLIST(browse_tab.ctree), 0, 1, results);
	   d_free(results);

	   conn_info.state &= ~BROWSING_MASK;
	   browse_results = 0;
	}
      break;
    case NAPSTER_BROWSE_RESULT:
	{
	   Search *browse;
	   
	   browse = gnapster_parse_browse(data, &data, 0);
	   if (!browse) 
	     break;

	   browse_results++;
	   
	   browse_tab_add(browse);
	}
      break;
    case NAPSTER_END_BROWSE:
	{
	   char *results;
	   
	   gnapster_update_status(conn_info.status, 0);
	   
	   gtk_entry_set_text(GTK_ENTRY(browse_tab.entry), "");
	   gtk_entry_set_editable(GTK_ENTRY(browse_tab.entry), 1);
	   
	   gtk_ctree_sort_recursive(GTK_CTREE(browse_tab.ctree), NULL);
	   gtk_clist_thaw(GTK_CLIST(browse_tab.ctree));
	   
	   if (g_list_length(GTK_CLIST(browse_tab.ctree)->row_list) == 0)
	     gnapster_clist_append(browse_tab.ctree, browse_tab.vbox, NULL,
				   "No mp3s found.", "", "", "", "", NULL);
      
	   results = d_strdup_printf(DEBUG "%i", browse_results);
	   gtk_clist_set_text(GTK_CLIST(browse_tab.ctree), 0, 1, results);
	   d_free(results);
      
	   conn_info.state &= ~BROWSING_MASK;
	   browse_results = 0;
	}
      break;
    case NAPSTER_BROWSE_ERROR:
      hook_text_insert(CONSOLE, SYSTEM, "general_message", "%s",
		       "user is not online");
      
      gnapster_update_status(conn_info.status, 0);
      
      gtk_clist_thaw(GTK_CLIST(browse_tab.ctree));
      
      conn_info.state &= ~BROWSING_MASK;
      
      break;
    case NAPSTER_DOWNLOAD_REPLY:
	{
	   Transfer *download, *real;
	   int row = -1;
	   
	   download = gnapster_parse_download(data, 0);
	   if (!download) 
	     break;
	   
	   real = gnapster_match_real_download(download->user, download->file, &row);
	   if (!real || real->status != 0) {
	      transfer_free(download);
	      break;
	   }
	   
	   real->ttype = DOWNLOAD;
	      
	   real->status++;
	   
      	   if (download->port == 0) {
	      napster_send(conn_info.sock, NAPSTER_ALT_GET, "%s \"%s\"",
			   download->user, download->file);
	      
	      gtk_clist_set_text(GTK_CLIST(download_tab.clist), row, 4,
				 "Awaiting connection...");	      
	   } else {
	      d_free(real->file);
	      d_free(real->user);
	      
	      push_transfer_info(real, download->user, download->file, NULL);

	      real->ip = download->ip;
	      real->port = download->port;
	      
	      gtk_clist_set_text(GTK_CLIST(download_tab.clist),
				 row, 4, "Connecting...");
	      
	      napster_connect_user(real);
	   }

	   transfer_free(download);
	}
      break;
    case NAPSTER_ALT_DOWNLOAD_REPLY:
	{
	   Transfer *upload;
	   
	   upload = gnapster_parse_download(data, 1);
	   if (!upload) 
	     break;
	   
	   /* Ok, we need to connect to them now */
	   upload->ttype = UPLOAD;
	   
	   if (upload->port == 0) {
	      hook_text_insert(CONSOLE, SYSTEM, "reject_upload", "%s\4%s",
			       upload->user, upload->trunc_file);
	      transfer_free(upload);
	   } else
	     napster_connect_user(upload);
	}
      break;
    case NAPSTER_DOWNLOAD_ERROR:
	{
	   char *user, *file;
	   Transfer *real;
	   
	   user = next_arg(data, &data);
	   file = next_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();

	   hook_text_insert(CONSOLE, SYSTEM, "whois_failure", "%s",
			    user);
	   real = gnapster_match_real_download(user, file, NULL);
	   download_close(real, TRANSFER_ERROR);
	}
      break;
    case NAPSTER_SENDLIMIT_REACHED:
	{
	   char *user, *file, *trunc_file;
	   unsigned long int size;
	   int limit, row = -1;
	   Transfer *real;
	   
	   user = next_arg(data, &data);
	   file = next_arg(data, &data);
	   convert(next_arg(data, &data), "%lu", &size);
	   limit = my_atoi(next_arg(data, &data));
	   
	   NA_ERR_HANDLE_BREAK();

	   trunc_file = strrchr(file, '\\');
	   if (!trunc_file) 
	     break;
	   
	   trunc_file++;
	   
	   real = gnapster_match_real_download(user, file, &row);
	   if (!real)
	     break;
	   
	   gtk_clist_set_text(GTK_CLIST(download_tab.clist), row, 4,
			      "Remotely queued");
	   
	   /* Try again every minute */
	   if (!real->timeout_tag && !user_info.conf[AUTOCANCEL]) {
	      if (real->start_timeout)
		gtk_timeout_remove(real->start_timeout);
	      
	      real->timeout_tag = gtk_timeout_add(60000 * 1,
						  remote_queue_check, real);
	   }
	   
	   real->status = 1;
	   
	   if (user_info.conf[AUTOCANCEL])
	     download_close(real, TRANSFER_ERROR);
	}
      break;
    case NAPSTER_PING:
	{
	   char *user;
	   
	   user = next_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   if (user_ignored(user))
	     break;
	   
	   hook_text_insert(CONSOLE, SYSTEM, "remote_ping", "%s",
			    user);
	   napster_send(conn_info.sock, NAPSTER_PONG, user);
	}
      break;
    case NAPSTER_ACTION:
	{
	   char *chan, *user, *message;
	   
	   chan = next_arg(data, &data);
	   user = next_arg(data, &data);
	   message = next_arg(data, &data);
	   
	   NA_ERR_HANDLE_BREAK();
	   
	   if (user_ignored(user) ||
	       !strcmp(user, user_info.user))
	     break;

	   hook_text_insert(chan_lookup(chan), SYSTEM, "public_action",
			    "%s\4%s", user, message);
	}
      break;
    case NAPSTER_REQUEST_PORTCHANGE:
	{
	   int port;
	   
	   port = my_atoi(next_arg(data, &data));
	   
	   NA_ERR_HANDLE_BREAK();

	   hook_text_insert(CONSOLE, SYSTEM, "general_message", "%s %i",
			    "server is attempting to change your port to",
			    port);
	   
	   handle_rebind(port);
	}
      break;
    case NAPSTER_NEWUSER_SUCCESS:
      napster_send(conn_info.sock, NAPSTER_NEW_USER_LOGIN, "%s %s %i \"%s\" %i %s",
		   user_info.user, user_info.pass, conn_info.nap_port, NAP_VERSION,
		   user_info.connection, "anonymous@napster.com");
      break;
    case NAPSTER_NEWUSER_REG:
      j_error_dialog(_("The username you have chosen has already been taken on napster.  Please choose a new one."));
      disconnect_cb();
      break;
    case NAPSTER_NEWUSER_INVALID_NICK:
      j_error_dialog(_("Username considered invalid by server.  Please try again"));
      disconnect_cb();
      break;
    case NAPSTER_ERROR:
      if (!strcmp(data, "Invalid Password"))
	j_error_dialog("Invalid Password.  This can be caused by attempting to connect with a user\nnot yet registered to the service.  Try checking New Account before attempting\nto do this and if the same problem occurs, then select a different user.");
      else if (!strncmp(data, "Both you and ", 13))
	break;
      else
	hook_text_insert(CURR, SYSTEM, "general_error", "%s", data);
      
      fprintf(stderr, "error -> %s\n", data);
      disconnect_cb();
      break;
    default:
      fprintf(stderr, "unknown[%i] -> %s\n", c, data);
      break;
   }
}

void napster_get_best_host() {
   d_free(user_info.server);
   user_info.server = d_config_get_string("/gnapster/Options/server");
   user_info.server_port = j_config_get_int("/gnapster/Options/server_port");
   
   dispatch_connect(inet_addr(user_info.server), htons(user_info.server_port),
		    NULL, &(conn_info.sock_input),
		    GDK_INPUT_WRITE, conn_main, NULL);
}

void napster_connect(char *ip, int port) {
   dispatch_connect(inet_addr(ip), htons(port),
		    &(conn_info.sock), &(conn_info.sock_input),
		    GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION,
		    conn_final, NULL);
}

void napster_connect_user(Transfer *t) {
   t->size = 0;

   dispatch_connect(t->ip, htons(t->port), &(t->sock), &(t->sock_input),
		    GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION,
		    conn_outgoing, t);
}

void handle_rebind(unsigned short nport) {
   if (user_info.conf[FIREWALL])
     nport = 0;
   
   if (conn_info.nap_port == nport)
     return;
   
   napster_unbind();
   napster_bind(nport);
}

void napster_unbind() {
   INPUT_REMOVE(conn_info.bind_sock_input);
   
   CLOSE(conn_info.bind_sock);
}

void napster_bind(unsigned short port) {
   int sock, len = 1;
   struct sockaddr_in server;
   
   conn_info.nap_port = port;
   
   if (!port)
     return;
   
   sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
   if (sock < 0) {
      j_error("socket", NULL);
      return;
   }
   
   memset(&server, 0, sizeof(server));
   server.sin_family = AF_INET;
   server.sin_addr.s_addr = htonl(INADDR_ANY);
   server.sin_port = htons(port);
   
   setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &len, sizeof(len));
   
   if (bind(sock, (struct sockaddr *)&server, sizeof(server)) < 0) {
      j_error("bind", NULL);
      j_error_dialog(_("An error occured while attempting to bind the configured port.  It is possible that another napster client is running."));
      close(sock);
      
      return;
   }
   
   listen(sock, 10 + user_info.conf[MAX_UPLOADS]);
   
   conn_info.bind_sock = sock;
   conn_info.bind_sock_input = gdk_input_add(sock, GDK_INPUT_READ,
					     handle_connection, NULL);
   
   return;
}

int napster_conn_as_int(char *s) {
   int x = 0;
   
   while(conn_table[x]) {
      if (!strcmp(conn_table[x], s)) return x;
      x++;
   }
   
   return -1;
}

char *napster_conn_as_str(int i) {
   if (i < 0 || i > 10) return "Unknown";
   return conn_table[i];
}

int napster_user_uploads(char *user) {
   GList *ptr;
   GtkCList *clist;
   GtkCListRow *clist_row;
   Transfer *data;
   int uploads = 0;
   
   clist = GTK_CLIST(upload_tab.clist);
   ptr = clist->row_list;
   if (!ptr)
     return -1;
   while(ptr) {
      clist_row = ptr->data;
      data = clist_row->data;
      if (!data)
	break;
      if (!strcmp(data->user, user))
	uploads++;
      ptr = ptr->next;
   }
   
   return uploads;
}
