#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <ctype.h>

#include <time.h>

#include <string.h>

#include "napster.h"
#include "gnapster.h"
#include "callbacks.h"
#include "ui.h"
#include "upload.h"
#include "network.h"
#include "commands.h"
#include "chan.h"
#include "xtext.h"

#include "utils.h"

#include "debug.h"

#undef read_file
#undef read_file_with_n

extern ConnInfo conn_info;
extern UserInfo user_info;
extern GList *hooks;
extern GList *total_list;

GList *log_hooks = NULL;

extern int na_err;

FreeTable gnapster_free[] = {
  /* Type               Function         Size */
   { N_SEARCH,		search_free,     sizeof(Search) },
   { TRANSFER,	        transfer_free,   sizeof(Transfer) },
   { N_SERVER,		server_free,     sizeof(Server) },
   { N_USER,		user_free,       sizeof(User) },
   { N_CHAN_INFO,       channel_free,    sizeof(ChannelInfo) },
   { N_RESUME,		resume_free,     sizeof(Resume) },
   { W_PROP,		property_free,   sizeof(WidgetProperty) },
   { HOOK,              hook_free,       sizeof(ThemeHook) },
   { THEME,             theme_free,      sizeof(Theme) },
   { FILE_ENTRY,        file_entry_free, sizeof(FileEntry) },
   { -1,		NULL,           -1 },
};

char *true_id(char *id) {
   char *pos;
   
   if ((pos = strchr(id, '-')))
     *pos = 0;
   
   return d_strdup(id);
}

void j_error(char *func, ...) {
   va_list args;
   char *arg, *prev;
   
   if (!strcmp(func, "fopen"))
     return;
   
   fprintf(stderr, "%s(", func);
   
   arg = prev = NULL;
   
   va_start(args, func);
   for(;;) {
      arg = va_arg(args, char *);
      if (!arg)
	break;
      
      fprintf(stderr, "%s%s", (prev) ? "," : "", arg);
      
      prev = arg;
   }
   va_end(args);
   
   fprintf(stderr, "): %s\n", strerror(errno));
}

void print_bits(int x) {
   int i;
   
   for(i=0; i < 32; i++) {
      printf("%i", (x & (1 << (31 - i))) ? 1 : 0);
      if (i % 4 == 3)
	printf(" ");
   }
   printf("\n");
}

char *trunc_file(char *path) {
   char *file;
   
   if (!path)
     return NULL;

   file = strrchr(path, '\\');
   if (!file)
     return path;
   
   /* push past the '\' and return */
   return (file + 1);
}

GList *remove_nick(GList *nick_history, char *user) {
   GList *ptr, *link, *ret;
   char *dptr;
   
   link = NULL;
   
   for(ptr=nick_history; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      if (!dptr)
	continue;
      
      if (!strcasecmp(dptr, user))
	link = ptr;
      
      if (link)
	break;
   }
   
   if (link) {
      d_free(link->data);
      link->data = NULL;
      
      ret = g_list_remove_link(nick_history, link);
      return ret;
   }
   
   return nick_history;
}

int cancels() { 
   return ((conn_info.state & CANCELS_MASK) >> 8);
}

int cancels_inc(int inc) {
   unsigned char c;
   
   c = (conn_info.state & CANCELS_MASK) >> 8;
   
   c += inc;
   
   conn_info.state &= ~CANCELS_MASK;
   conn_info.state |= (c << 8);
   
   return c;
}

int connected() {
   return (conn_info.state & CONNECTED_MASK);
}

int connecting() {
   return (conn_info.state & CONNECTING_MASK);
}

int searching() {
   return (conn_info.state & SEARCHING_MASK);
}

int browsing() {
   return (conn_info.state & BROWSING_MASK);
}

int safe_exit() {
   return (conn_info.state & SAFE_EXIT_MASK);
}

char *global_path(char *subdir, char *file) {
   char *path;
   
   if (file)
     path = d_strdup_printf(DEBUG "%s/%s/%s", CONFIG_PATH, subdir, file);
   else
     path = d_strdup_printf(DEBUG "%s/%s", CONFIG_PATH, subdir);
   
   return path;
}

char *local_path(char *subdir, char *file) {
   char *path;
   
   if (file)
     path = d_strdup_printf(DEBUG "%s/.gnapster/%s/%s", 
			    getenv("HOME"), subdir, file);
   else
     path = d_strdup_printf(DEBUG "%s/.gnapster/%s",
			    getenv("HOME"), subdir);
   
   return path;
}

int is_comment(char *text) {
   char *ptr;
   
   ptr = text;
   while(*ptr == ' ')
     ptr++;
  
   return (*ptr == '#');
}

char *hack_n(char *s) {
   return hack_chr(s, '\n');
}

char *hack_chr(char *s, char c) {
   char *ptr;
   
   if (!s)
     return NULL;
   
   ptr = s;
   
   while(*ptr) {
      if (*ptr == c) {
	*ptr = 0;
	break;
      }
      
      ptr++;
   }
   
   return s;
}

void read_file_free(GList *list) {
   GList *ptr;
   FileEntry *fent;
   
   for(ptr=list; ptr; ptr=ptr->next) {
      fent = ptr->data;
      if (!fent)
	continue;
      
      file_entry_free(fent);
      ptr->data = NULL;
   }
   
   g_list_free(list);
}

GList *read_file_full(const char *file, int line, FILE *f, int allow_n) {
   char buf[4096];
   GList *list = NULL;
   FileEntry *fent;
   
   for(;;) {
      fgets(buf, sizeof(buf) - 1, f);
      if (feof(f)) 
	break;
      
      if (!buf[0])
	continue;
      
      if (allow_n || buf[0] != '\n') {
	 fent = d_new(FILE_ENTRY);
	 
	 /* this is actually setup like this so next arg can be used on
	  * fent->lptr, and keep fent->line at the right memory address */
	 fent->line = d_strdup(buf);
	 fent->lptr = fent->line;
	 
	 list = g_list_append(list, fent);
      }
   }

   return list;
}

GList *read_file(const char *file, int line, FILE *f) {
   GList *list;
   
   list = read_file_full(file, line, f, 0);
   
   return list;
}

GList *read_file_with_n(const char *file, int line, FILE *f) {
   GList *list;
   
   list = read_file_full(file, line, f, 1);
   
   return list;
}

size_t struct_size(int t) {
   int i;
   
   for(i=0; gnapster_free[i].func; i++) 
     if (gnapster_free[i].type == t)
       return gnapster_free[i].size;
   
   return -1;
}

int get_type(void *data) {
   int t = -1;

   if (data)
     t = ((Type *)data)->type;

   return t;
}

void set_type(void *data, int t) {
   if (t < 0 || !data)
     return;
   
   ((Type *)data)->type = t;
}

char *all_lower(char *s) {
   int x;
   static char ret[1024 * 6];
   
   if (!s)
     return NULL;

   x = strlen(s);
   if (x <= 0)
     return NULL;

   if (x > sizeof(ret))
     return s;

   memset(ret, 0, sizeof(ret));

   while(x--)
     ret[x] = tolower(s[x]);

   return ret;
}

void conv_file(char *s) {
   char *ptr;

   if (!user_info.conf[NAME_FIX])
     return;

   ptr = strrchr(s, '/');
   if (!ptr)
     return;
   while(*ptr) {
      if (*ptr == ' ') *ptr = '_';
      ptr++;
   }
}

int file_exists(char *file) {
   struct stat st;
   int x;

   x = stat(file, &st);

   return (x >= 0);
}

char *next_arg_full_max_length(char *s, char **next, char end, int max) {
   char *ptr, *str;
   
   na_err = (!s || !(*s));
   if (na_err)
     return NULL;

   ptr = s;
   
   str = (end == '"') ? 
     ++ptr : ptr;
   
   while(*str && *str != end && *str != '\n' && max)
     str++, max--;
   
   if (*str) {
      *str++ = 0;
      if (end == '"') str++;
   }

   if (next)
     *next = str;

   return ptr;
}

char *next_arg_full(char *s, char **next, char end) {
   return next_arg_full_max_length(s, next, end, -1);
}

char *next_arg_first(char *s, char **next, char **first) {
   *first = NULL;

   if (!s || !*s) {
      na_err = 1;
      return NULL;
   }

   *first = s;

   return next_arg(s, next);
}

char *next_arg(char *s, char **next) {
   char cmd;

   if (!s || !*s) {
      na_err = 1;
      return NULL;
   }

   cmd = (*s == '"') ? '"' : ' ';

   return next_arg_full(s, next, cmd);
}

char *last_arg(char *s, char **next) {
   return next_arg_full(s, next, 0);
}

void convert(char *s, char *type, void *c) {
   char *ptr;
   
   if (!s)
     return;
   
   ptr = s;
   
   while(*ptr == '0' && *(ptr + 1) != 0)
     ptr++;

   if (ptr)
     sscanf(ptr, type, c);
}
      
long my_atol(const char *nptr) {
   if (!nptr) return -1;

   return atol(nptr);
}

int my_atoi(const char *nptr) {
   if (!nptr) return -1;
   
   return atoi(nptr);
}

char *simplify_size(unsigned long int size) {
   float new_size;
   int div_count;
   static char buf[25] = { 0 };
   char lbl, div_table[4] = { 'b', 'k', 'M', ' ' };
   
   new_size = size;
   div_count = 0;
   
   while(new_size >= 1024) {
      new_size /= 1024;
      div_count++;
   }
   
   lbl = div_table[div_count];

   g_snprintf(buf, sizeof(buf), "%.02f%c", new_size, lbl);
   
   return buf;
}

void mp3_list_free() {
   GList *ptr, *child, *cptr;
   MP3Stats *mp3;
   
   if (!total_list)
     return;
   
   for(ptr=total_list; ptr; ptr=ptr->next) {
      child = ptr->data;
      if (!child)
	continue;
      
      if (child->data)
	d_free(child->data);
      
      child->data = NULL;
      
      for(cptr=child->next; cptr; cptr=cptr->next) {
	 mp3 = cptr->data;
	 if (mp3)
	   mp3_free(mp3);
	 
	 cptr->data = NULL;

//	 total_list = g_list_remove_link(total_list, cptr);
      }
      
      g_list_free(child);
   }
   
   g_list_free(total_list);
   
   total_list = NULL;
}

void file_entry_free(void *data) {
   FileEntry *fent;
   
   if (!data)
     return;
   
   fent = (FileEntry *)data;
   
   d_free(fent->line);
   d_free(fent);
}

void theme_free(void *data) {
   Theme *theme;
   
   if (!data)
     return;
   
   theme = (Theme *)data;
   
   d_free(theme->file);
   d_free(theme->name);
   d_free(theme->author);
   d_free(theme);
}

void hook_free(void *data) {
   ThemeHook *hook;
   
   if (!data)
     return;
   
   hook = (ThemeHook *)data;
   
   d_free(hook->name);
   d_free(hook->data);
   d_free(hook);
}

void property_free(void *data) {
   WidgetProperty *prop;

   if (!data) 
     return;
   
   prop = (WidgetProperty *)data;
   
   d_free(prop->conf);
   g_list_free(prop->sensitive);
   d_free(prop);
}

void transfer_free(void *data) {
   Transfer *t;
   
   if (!data) 
     return;
   
   t = data;
   
//   d_free(t->trunc_file);
   d_free(t->file);
   d_free(t->header);
   d_free(t->id);
   d_free(t->user);
   d_free(t->fpath);
   d_free(t->resume_path);
   d_free(t);
}

void mp3_free(void *data) {
   MP3Stats *mp3;
   
   if (!data)
     return;
   
   mp3 = data;
   
   d_free(mp3->filename);
   d_free(mp3->checksum);
   d_free(mp3);
}

void search_free(void *data) {
   Search *search;
   
   if (!data) 
     return;
   
   search = (Search *)data;
   
//   d_free(search->trunc_file);
   d_free(search->file);
   d_free(search->user);
   d_free(search->id);
   d_free(search);
}

void server_free(void *data) {
   Server *server;
   
   if (!data) 
     return;
   
   server = (Server *)data;
   
   d_free(server->ip);
   d_free(server->desc);
   d_free(server->network);
   d_free(server);
}

void user_free(void *data) {
   User *user;
   
   if (!data) 
     return;
   
   user = (User *)data;
   
   d_free(user->user);
   d_free(user);
}

void channel_free(void *data) {
   Channel *channel;
   
   if (!data) 
     return;
   
   channel = (Channel *)data;
   
   d_free(channel->chan);
   d_free(channel->topic);
   d_free(channel);
}

void resume_free(void *data) {
   Resume *resume;
   
   if (!data) 
     return;
   
   resume = (Resume *)data;
   
   d_free(resume->file);
   d_free(resume->checksum);
   d_free(resume);
}

void gnapster_destroy(gpointer data) {
   int i = 0, type;
   
   if (!data) 
     return;
   
   type = get_type(data);
   
   while(gnapster_free[i].func) {
      if (type == gnapster_free[i].type) {
	 gnapster_free[i].func (data);
	 break;
      }
      i++;
   }
   
   if (!gnapster_free[i].func)
     d_free(data);
}

int same_color(GdkColor col1, GdkColor col2) {
   if (col1.red == col2.red &&
       col1.green == col2.green &&
       col1.blue == col2.blue)
     return 1;
   
   return 0;
}

void unlink_log() {
   char *path;
   
   path = local_path("logs", "messages");
//   if (file_exists(path))
//     unlink(path);
   
   d_free(path);
}

int copy_file(char *opath, char *npath) {
   char buf[2048];
   int o_fd, n_fd;
   int n;
   
   o_fd = open(opath, O_RDONLY);
   if (o_fd < 0)
     return o_fd;

   n_fd = open(npath, O_CREAT | O_WRONLY | O_TRUNC, 0644);
   if (n_fd < 0) {
      close(o_fd);
      return n_fd;
   }
   
   while((n = read(o_fd, buf, sizeof(buf))) > 0)
     write(n_fd, buf, n);
   
   close(o_fd);
   close(n_fd);
   
   return 1;
}

/* this function copies your ~/.gnome/gnapster file to 
 * ~/.gnapster/gnapster.conf if you do not have GNOME enabled 
 * 
 * it also checks to make sure that ~/.gnapter/gnapster.conf doesn't
 * already exist */
void relocate_config(char *home) {
#ifndef HAVE_GNOME
   char *orig_path, *new_path;
   
   orig_path = d_strdup_printf(DEBUG "%s/.gnome/gnapster", home);
   
   if (!file_exists(orig_path)) {
      d_free(orig_path);
      
      return;
   }
   
   new_path = local_path("gnapster.conf", NULL);
   
   if (file_exists(new_path)) {
      d_free(orig_path);
      d_free(new_path);
      
      return;
   }
   
   copy_file(orig_path, new_path);
   
   d_free(orig_path);
   d_free(new_path);
#endif /* !HAVE_GNOME */
}

void create_priv_dir() {
   FILE *aj_f;
   struct stat st;
   char *home, *path;
   
   home = getenv("HOME");
   
   relocate_config(home);
   
   path = local_path("", NULL);
   if (stat(path, &st) < 0)
     if (mkdir(path, 0755) < 0)
       fprintf(stderr, "Could not create ~/.gnapster directory...functionality may be limited\n");
   
   d_free(path);
   
   path = local_path("logs", NULL);
   if (stat(path, &st) < 0)
     if (mkdir(path, 0755) < 0)
       fprintf(stderr, "Could not create ~/.gnapster/logs directory...\n");
   
   d_free(path);
   
   path = local_path("themes", NULL);
   if (stat(path, &st) < 0)
     if (mkdir(path, 0755) < 0)
       fprintf(stderr, "Could not create ~/.gnapster/themes directory...\n");
   
   d_free(path);
   
   if (!(aj_f = open_autojoin("r"))) {
      FILE *f;
      
      f = open_autojoin("w");
      if (!f)
	return;

      fprintf(stderr, "autojoin not found, creating defaults...\n");

      fprintf(f, "#gnapster\n");
      
      fclose(f);
   } else
     fclose(aj_f);
}

FILE *open_log(char *file, char *flags) {
   FILE *f;
   char *conf;
   
   conf = local_path("logs", file);
   
   f = fopen(conf, flags);
   
   d_free(conf);
   
   return f;
}

void log_add_hooks(char *place, ...) {
   va_list args;
   char *hook;
   
   va_start(args, place);
   for(;;) {
      hook = va_arg(args, char *);
      if (!hook)
	break;
      
      log_hooks = g_list_append(log_hooks, d_strdup(hook));
   }
   va_end(args);
}

int hook_log(char *hook) {
   GList *ptr;
   char *dptr;
   
   if (!hook)
     return 0;
   
   for(ptr=log_hooks; ptr; ptr=ptr->next) {
      dptr = ptr->data;
      if (!dptr)
	continue;
      
      if (!strcmp(hook, dptr))
	return 1;
   }
   
   return 0;
}

int match_date(char *path, FILE *f) {
   char buf[11];
   FILE *new_file;
   struct tm *tm;
   char *dptr;
   time_t timep;
   int month, day, year, fd, n;
   
   timep = time(NULL);
   tm = localtime((time_t *)&timep);
   
   if (!file_exists(path)) {
      new_file = fopen(path, "w");
      fprintf(new_file, "%02i/%02i/%04i\n\n",
	      tm->tm_mon + 1, tm->tm_mday, 1900 + tm->tm_year);
      fclose(new_file);
   }
   
   fd = open(path, O_RDONLY);
   if (fd < 0)
     return 0;

   n = read(fd, buf, sizeof(buf) - 1);
   if (n < 0) {
      close(fd);
      return 0;
   }
   
   buf[n] = 0;
   
   dptr = buf;
   convert(next_arg_full(dptr, &dptr, '/'), "%i", &month);
   convert(next_arg_full(dptr, &dptr, '/'), "%i", &day);
   convert(next_arg(dptr, &dptr), "%i", &year);
   
   close(fd);
   
   if (na_err) {
      new_file = fopen(path, "w");
      fprintf(new_file, "%02i/%02i/%04i\n\n",
	      tm->tm_mon + 1, tm->tm_mday, 1900 + tm->tm_year);
      fclose(new_file);
   }
   
   NA_ERR_HANDLE(1);

   /* was it created today? */
   if (tm->tm_mday == day && (tm->tm_mon + 1) == month && (1900 + tm->tm_year) == year)
     return 1;
   
   return 0;
}

void move_file(char *path, FILE **f) {
   char *npath;
   
   npath = d_strdup_printf(DEBUG "%s.old", path);
   
   rename(path, npath);
   
   d_free(npath);
}

void rotate_logs(char *path, FILE **f) {
   if (!f)
     return;

   if (!match_date(path, *f))
     move_file(path, f);
   
   *f = fopen(path, "a");

   d_free(path);
}

void console_log(char *hook, char *text) {
   FILE *f;
   char *stripped;
   
   if (!user_info.conf[LOGGING] || !hook || !text)
     return;
   
   if (!hook_log(hook))
     return;
   
   rotate_logs(local_path("logs", "messages"), &f);
   
   if (!f)
     return;
   
   stripped = gtk_xtext_strip_color(text, strlen(text), NULL, NULL);
   if (stripped) {
      d_memory_hunk_append(DEBUG stripped, strlen(stripped));
      
      /* \n is already on there */
      fprintf(f, "%s", stripped);
   }
   
   d_free(stripped);
   
   fclose(f);
}
